"""Build an approximation of a single-key PCB.

<span class="todo">TODO: Enapsulate and document this better.</span>
"""
from solid2 import cube, cylinder, hull, mirror, rotate, up, down, left, right, forward, back, scad_render_to_file
from solid2.core.object_base import OpenSCADObject

from .switch_plate import (
    keyswitch_depth,
    plate_thickness,
)
from .utils import cylinder_outer


pcb_thickness = 1.57
pcb_size = 19
pcb_corner_radius = 1.25
pcb_corner_offset = pcb_size / 2 - pcb_corner_radius
pcb_hole_radius = 1.7 / 2
pcb_hole_offset = 8


def single_key_board(simple: bool = False) -> OpenSCADObject:
    """Build an approximation of a single-key PCB.

    If `simple=True`, then just approximate the size of one of these with a rectangular prism.

    Some compatible single-key PCBs:
    - https://www.flux.ai/whitelynx/mx-single-keyswitch-hot-swap-board
    - https://www.flux.ai/whitelynx/choc-single-keyswitch-hot-swap-board
    """
    if simple:
        pcb = cube((pcb_size, pcb_size, pcb_thickness), center=True)
    else:
        pcb_corner = cylinder_outer(r=pcb_corner_radius, h=pcb_thickness, center=True)
        pcb_hole = cylinder_outer(r=pcb_hole_radius, h=pcb_thickness * 2, center=True)

        pcb = (
            hull()(
                pcb_corner.left(pcb_corner_offset).forward(pcb_corner_offset),
                pcb_corner.left(pcb_corner_offset).back(pcb_corner_offset),
                pcb_corner.right(pcb_corner_offset).forward(pcb_corner_offset),
                pcb_corner.right(pcb_corner_offset).back(pcb_corner_offset),
            )
            - pcb_hole.left(pcb_hole_offset).forward(pcb_hole_offset)
            - pcb_hole.left(pcb_hole_offset).back(pcb_hole_offset)
            - pcb_hole.right(pcb_hole_offset).forward(pcb_hole_offset)
            - pcb_hole.right(pcb_hole_offset).back(pcb_hole_offset)
        )

    return pcb.down(pcb_thickness / 2 - plate_thickness + keyswitch_depth)


# To test, use the command line: pipenv run python -m spkb.single_key_pcb
if __name__ == "__main__":
    print("Rendering single_key_board() to single_key_board.scad...")
    single_key_board().save_as_scad("single_key_board.scad")



from typing import Optional

from solid2 import rotate, cube, hull, scad_render_to_file, up, left, right, forward, back
from solid2.core.object_base import OpenSCADObject

from .utils import cylinder_outer, optional


SEGMENTS = 48

FUDGE = 0.2

m2_head_radius = 5 / 2
m2_shaft_radius = 2 / 2
m2_nut_radius = 3.9 / 2

mount_post_m2_radius = 6 / 2


def mount_post_m2(height) -> OpenSCADObject:
    return (
        cylinder_outer(mount_post_m2_radius, height)
        - cylinder_outer(m2_shaft_radius, height + 0.1).down(0.05)
    )


class BoardMount:
    def __init__(self,
                 board_width: float,
                 board_length: float,
                 board_thickness: float,
                 has_connector: bool = True,
                 front_mounting_post_separation: float = 10,
                 back_mounting_post_separation: Optional[float] = None,
                 ):
        self.board_width = board_width
        self.board_length = board_length
        self.board_thickness = board_thickness
        self.has_connector = has_connector

        self.front_mounting_post_separation = front_mounting_post_separation
        self.back_mounting_post_separation = back_mounting_post_separation

        # Distance from the front edge of the board to the front of the large portion of the plug
        self.plug_offset = 2

        # Length of the plug clearance volume
        self.plug_length = 20

    def pcb_only(self, distance_from_surface: float) -> OpenSCADObject:
        return up(distance_from_surface + self.board_thickness / 2)(
            back(self.board_length / 2)(
                cube(
                    (self.board_width, self.board_length, self.board_thickness),
                    center=True,
                )
            )
        )

    def connector(self, distance_from_surface: float) -> OpenSCADObject:
        """An approximation of a USB-C connector.
        """
        return optional(self.has_connector)(
            up(distance_from_surface - 1.25)(
                forward(self.plug_offset + 0.1)(
                    rotate((90, 0, 0))(
                        hull()(
                            left(4 - 1.25)(cylinder_outer(2.5 / 2, 6.2)),
                            right(4 - 1.25)(cylinder_outer(2.5 / 2, 6.2)),
                        )
                    )
                )
                + forward(self.plug_offset + self.plug_length)(
                    rotate((90, 0, 0))(
                        hull()(
                            left(4 - 1.25)(cylinder_outer(8.5 / 2, self.plug_length)),
                            right(4 - 1.25)(cylinder_outer(8.5 / 2, self.plug_length)),
                        )
                    )
                )
            )
        )

    def board_profile(self, distance_from_surface: float) -> OpenSCADObject:
        return (
            self.pcb_only(distance_from_surface)
            + self.connector(distance_from_surface)
        )
        # TODO: Maybe add pin clearance!

    def back_mounting_posts(self, distance_from_surface: float) -> OpenSCADObject:
        mounting_post = mount_post_m2(distance_from_surface)

        mounting_post_shift = mount_post_m2_radius * 2
        if self.back_mounting_post_separation is not None:
            mounting_post_shift = mount_post_m2_radius + self.back_mounting_post_separation / 2

        return back(self.board_length + m2_shaft_radius + FUDGE)(
            mount_post_m2(distance_from_surface)
            if self.back_mounting_post_separation is None else (
                mount_post_m2(distance_from_surface).left(mounting_post_shift)
                + mount_post_m2(distance_from_surface).right(mounting_post_shift)
            )
        )

    def front_mounting_posts(self, distance_from_surface: float) -> OpenSCADObject:
        positioning_post_height = distance_from_surface + self.board_thickness + 3
        positioning_post = cube((4, 4, positioning_post_height), center=True) \
            .up(positioning_post_height / 2) \
            .forward(1)

        positioning_post_shift = 2 + self.front_mounting_post_separation / 2

        return (
            left(positioning_post_shift)(positioning_post)
            + right(positioning_post_shift)(positioning_post)
        ) - self.board_profile(distance_from_surface)

    def mounting_posts(self, distance_from_surface: float) -> OpenSCADObject:
        return (
            self.back_mounting_posts(distance_from_surface)
            + self.front_mounting_posts(distance_from_surface)
        )

    def render(self, distance_from_surface: float) -> OpenSCADObject:
        return self.mounting_posts(distance_from_surface) - self.board_profile(
            distance_from_surface
        )


pro_micro = BoardMount(18.3, 33.1, 1.7)
stm32_blackpill = BoardMount(20.66, 53, 1.64, back_mounting_post_separation=11)


# To test, use the command line: pipenv run python -m spkb.board_mount
if __name__ == "__main__":
    print("Rendering pro_micro to pro_micro.scad...")
    pro_micro.save_as_scad("pro_micro.scad")

    print("Rendering stm32_blackpill to stm32_blackpill.scad...")
    stm32_blackpill.save_as_scad("stm32_blackpill.scad")



"""Renderers for a variety of keyswitch plates (sockets)

<span markdown="1" class="deprecated">Deprecated: Use classes in `spkb.keyswitch` instead.</span>
"""
from solid2 import cube, cylinder, hull, mirror, rotate, up, down, left, right, forward, back, scad_render_to_file
from solid2.core.object_base import OpenSCADObject


keyswitch_length = 14.0
keyswitch_width = 14.0
keyswitch_depth = 5.08  # From the base of the switch to the mounting plate face

plate_thickness = 3
notch_plate_thickness = 1.3  # The thickness of the plate at the notches where the switch's clips are located
notch_width = 5
notch_width_outer = 6
notch_depth = 0.5
notch_height = 8
notch_height_outer = 9
backplate_thickness = 1.25
backplate_orientation = 180
mount_width = keyswitch_width + 3
mount_length = keyswitch_length + 3

cherry_backplate_clearance_distance = 3.5


def mx_plate(full_depth: bool = False, extra_depth: float = 0, wall_thickness: float = 1.5) -> OpenSCADObject:
    """Build a segment of plate for an MX-compatible switch.
    """
    thickness = (keyswitch_depth if full_depth else plate_thickness) + extra_depth
    top_wall = forward((wall_thickness + keyswitch_length) / 2)(
        up(plate_thickness - thickness / 2)(
            cube((keyswitch_width + wall_thickness * 2, wall_thickness, thickness), center=True)
        )
        - down(notch_plate_thickness + notch_height / 2 - plate_thickness)(  # Notch for switch clips
            back(wall_thickness / 2)(
                hull()(
                    cube((notch_width, notch_depth * 2, notch_height), center=True),
                    back(notch_depth)(
                        cube((notch_width_outer, notch_depth * 2, notch_height_outer), center=True)
                    )
                )
            )
        )
    )

    left_wall = left((wall_thickness + keyswitch_width) / 2)(
        up(plate_thickness - thickness / 2)(
            cube((wall_thickness, keyswitch_length + wall_thickness * 2, thickness), center=True)
        )
    )

    plate_half = top_wall + left_wall

    return plate_half + mirror((0, 1, 0))(mirror((1, 0, 0))(plate_half))


def mx_plate_with_board_mount(extra_depth: float = 0) -> OpenSCADObject:
    """Build a segment of plate for an MX-compatible switch, with mounting holes for a single-key PCB.

    Some compatible single-key PCBs:
    - https://www.flux.ai/whitelynx/mx-single-keyswitch-hot-swap-board
    - https://www.flux.ai/whitelynx/choc-single-keyswitch-hot-swap-board
    """
    screw_hole = down(keyswitch_depth / 2)(
        cylinder(r=0.5, h=keyswitch_depth + plate_thickness / 2, center=True, _fn=16)
    )

    return (
        mx_plate(full_depth=True, wall_thickness=2.625)
        - forward(8)(right(8)(screw_hole))
        - forward(8)(left(8)(screw_hole))
        - back(8)(right(8)(screw_hole))
        - back(8)(left(8)(screw_hole))
    )


def mx_backplate() -> OpenSCADObject:
    return up(plate_thickness - keyswitch_depth - backplate_thickness / 2)(
        rotate(backplate_orientation, [0, 0, 1])(
            cube((keyswitch_width + 3, keyswitch_length + 3, backplate_thickness), center=True)
            - cylinder(r=1.9939, h=backplate_thickness + 1, center=True, _fn=16)
            - right(5.08)(cylinder(r=0.8509, h=backplate_thickness + 1, center=True, _fn=16))
            - left(5.08)(cylinder(r=0.8509, h=backplate_thickness + 1, center=True, _fn=16))
            - left(3.81)(forward(2.54)(cylinder(r=1.5, h=backplate_thickness + 1, center=True, _fn=16)))
            - right(2.54)(forward(5.08)(cylinder(r=1.5, h=backplate_thickness + 1, center=True, _fn=16)))
            - right(1.27)(back(5.08)(cylinder(r=0.4953, h=backplate_thickness + 1, center=True, _fn=16)))
            - left(1.27)(back(5.08)(cylinder(r=0.4953, h=backplate_thickness + 1, center=True, _fn=16)))
            - right(3.81)(back(5.08)(cylinder(r=0.4953, h=backplate_thickness + 1, center=True, _fn=16)))
            - left(3.81)(back(5.08)(cylinder(r=0.4953, h=backplate_thickness + 1, center=True, _fn=16)))
        )
    )


def mx_backplate_clearance() -> OpenSCADObject:
    return back(-7.5 / 2)(
        up(plate_thickness - keyswitch_depth - backplate_thickness - cherry_backplate_clearance_distance / 2)(
            rotate(backplate_orientation, [0, 0, 1])(
                cube((16.5, 7.5, cherry_backplate_clearance_distance))
            )
        )
    )


def mx_plate_with_backplate() -> OpenSCADObject:
    """Build a segment of plate for an MX-compatible switch, with a backplate capable of holding a hot-swap socket.
    """
    return mx_plate(full_depth=True, extra_depth=1) + mx_backplate()


switch_plate = mx_plate


# To test, use the command line: pipenv run python -m spkb.switch_plate
if __name__ == "__main__":
    print("Rendering mx_plate() to mx_plate.scad...")
    mx_plate().save_as_scad("mx_plate.scad")

    print("Rendering mx_plate_with_backplate() to mx_plate_with_backplate.scad...")
    mx_plate_with_backplate().save_as_scad("mx_plate_with_backplate.scad")

    print("Rendering mx_plate_with_board_mount() to mx_plate_with_board_mount.scad...")
    mx_plate_with_board_mount().save_as_scad("mx_plate_with_board_mount.scad")



"""Renders a single-switch tester.

<span class="todo">TODO: Document or remove this module.</span>
"""
from solid2 import cube, rotate, up, forward, scad_render_to_file
from solid2.core.object_base import OpenSCADObject

from .switch_plate import (
    switch_plate,
    keyswitch_depth,
    plate_thickness,
    mount_width,
    mount_length,
)


switch_spacing = 10

wall_height = keyswitch_depth + 15
wall_length = max(mount_width, mount_length) + 2 * switch_spacing
wall_thickness = 5


def spaced_switch_plate() -> OpenSCADObject:
    plate_spacer = up(plate_thickness / 2)(
        forward((max(mount_width, mount_length) + switch_spacing) / 2)(
            cube((wall_length, switch_spacing, plate_thickness), center=True)
        )
    )

    return (
        switch_plate()
        + plate_spacer
        + rotate(90)(plate_spacer)
        + rotate(180)(plate_spacer)
        + rotate(270)(plate_spacer)
    )


def single_tester_walls() -> OpenSCADObject:
    wall = up(wall_height / 2)(
        forward((wall_length - wall_thickness) / 2)(
            cube((wall_length, wall_thickness, wall_height), center=True)
        )
    )

    return wall + rotate(90)(wall) + rotate(180)(wall) + rotate(270)(wall)


def single_tester() -> OpenSCADObject:
    return single_tester_walls() + up(wall_height - plate_thickness)(
        spaced_switch_plate()
    )


# To test, use the command line: pipenv run python -m spkb.single_tester
if __name__ == "__main__":
    print("Rendering single_tester() to single_tester.scad...")
    single_tester().save_as_scad("single_tester.scad")



from itertools import chain
from math import fabs
from typing import List, Tuple

from solid2 import cube, hull, rotate
from solid2.core.object_base import OpenSCADObject

from ..utils import cylinder_outer

Offset2D = Tuple[float, float]


class Keyswitch:
    """An encapsulation of all measurements and shapes related to a given type of keyswitch.
    """
    keyswitch_length: float = 14.0
    "The length (front to back) of the keyswitch mounting hole"
    keyswitch_width: float = 14.0
    "The width (left to right) of the keyswitch mounting hole"
    keyswitch_depth: float = 5.08
    "From the base of the switch to the mounting plate face"

    plate_thickness: float = 3
    "The thickness of the plate above the origin"
    notch_plate_thickness: float = 1.3
    "The thickness of the plate at the notches where the switch's clips are located"
    notch_width: float = 5
    "The width of the notch for the switch's clips at the deepest part of the notch"
    notch_width_outer: float = 6
    "The width of the notch for the switch's clips at the edge of the mounting hole"
    notch_depth: float = 0.5
    "The depth of the notch for the switch's clips"
    notch_height: float = 8
    "The height of the notch for the switch's clips at the deepest part of the notch"
    notch_height_outer: float = 9
    "The height of the notch for the switch's clips at the edge of the mounting hole"

    wall_thickness: float = 1.5
    "The thickness of the mounting hole walls"

    switch_midline_width: float = 13.95
    "The width (left to right) of the switch at its midline (at the top of the plate)"
    switch_midline_length: float = 15.6
    "The length (front to back) of the switch at its midline (at the top of the plate)"
    switch_topline_width: float = 10.5
    "The width (left to right) of the switch at its top"
    switch_topline_length: float = 10.5
    "The length (front to back) of the switch at its top"
    switch_height_above_plate: float = 6.2
    "The height of the switch body above the top of the plate"

    def plate(
        self,
        full_depth: bool = False,
        extra_depth: float = 0,
        wall_thickness: float = None
    ) -> OpenSCADObject:
        """Build a segment of plate for mounting this type of switch.

        :param full_depth: If True, extend the walls of the socket to the full depth of the switch; otherwise, only
        extend to the plate thickness.
        :param extra_depth: Extra depth (`z` height) to add to the walls of the socket.
        :param wall_thickness: The thickness of the walls of the socket.
        """
        if wall_thickness is None:
            wall_thickness = self.wall_thickness

        thickness = (self.keyswitch_depth if full_depth else self.plate_thickness) + extra_depth

        top_wall = (
            cube((self.keyswitch_width + wall_thickness * 2, wall_thickness, thickness), center=True)
            .up(self.plate_thickness - thickness / 2)
            # Notch for switch clips
            - hull()(
                cube((self.notch_width, self.notch_depth * 2, self.notch_height), center=True),
                cube((self.notch_width_outer, self.notch_depth * 2, self.notch_height_outer), center=True)
                .back(self.notch_depth),
            )
            .back(wall_thickness / 2)
            .down(self.notch_plate_thickness + self.notch_height / 2 - self.plate_thickness)
        ).forward((wall_thickness + self.keyswitch_length) / 2)

        left_wall = (
            cube((wall_thickness, self.keyswitch_length + wall_thickness * 2, thickness), center=True)
            .up(self.plate_thickness - thickness / 2)
            .left((wall_thickness + self.keyswitch_width) / 2)
        )

        plate_half = top_wall + left_wall

        return plate_half + plate_half.rotate(180, [0, 0, 1])

    def plate_with_board_mount(
        self,
        screw_positions: List[Offset2D],
        screw_radius: float = 0.5,
        extra_depth: float = 0,
        wall_thickness: float = None
    ) -> OpenSCADObject:
        """Build a segment of plate for mounting this type of switch, with mounting holes for a single-key PCB.

        :param screw_positions: The positions (`(x, y)` tuples) of the screws for mounting the PCB.
        :param extra_depth: Extra depth (`z` height) to add to the walls of the socket.
        :param wall_thickness: The thickness of the walls of the socket.
        """
        screw_hole = down(self.keyswitch_depth / 2)(
            cylinder_outer(r=screw_radius, h=self.keyswitch_depth + self.plate_thickness / 2, center=True)
        )

        if wall_thickness is None:
            # Find the furthest screw center from the edges of the keyswitch mounting hole.
            max_screw_offset_from_hole = max(
                chain.from_iterable(
                    (fabs(screw_x) - self.keyswitch_width / 2, fabs(screw_y) - self.keyswitch_length / 2)
                    for screw_x, screw_y in screw_positions
                )
            )

            # Add an extra screw_radius outside the screw hole, and use that to determine our wall thickness.
            wall_thickness = max_screw_offset_from_hole + screw_radius

        plate = self.plate(full_depth=True, wall_thickness=wall_thickness)
        for screw_x, screw_y in screw_positions:
            plate -= screw_hole.right(screw_y).forward(screw_x)

        return plate

    def switch(self) -> OpenSCADObject:
        """Build an simplified approximation of (the top half of) an MX-style keyswitch.
        """
        return hull()(
            cube(self.switch_midline_width, self.switch_midline_length, 0.1, center=True)
            .up(3.05),
            cube(self.switch_topline_width, self.switch_topline_length, 0.1, center=True)
            .up(self.switch_height_above_plate + 2.95),
        )



from solid2 import cube
from solid2.core.object_base import OpenSCADObject

from ..utils import cylinder_outer
from .base import Keyswitch


class MX(Keyswitch):
    """An encapsulation of all measurements and shapes related to Cherry MX-style switches.
    """
    keyswitch_length: float = 14.0
    "The length (front to back) of the keyswitch mounting hole"
    keyswitch_width: float = 14.0
    "The width (left to right) of the keyswitch mounting hole"
    keyswitch_depth: float = 5.08
    "From the base of the switch to the mounting plate face"

    plate_thickness: float = 3
    "The thickness of the plate above the origin"
    notch_plate_thickness: float = 1.3
    "The thickness of the plate at the notches where the switch's clips are located"
    notch_width: float = 5
    "The width of the notch for the switch's clips at the deepest part of the notch"
    notch_width_outer: float = 6
    "The width of the notch for the switch's clips at the edge of the mounting hole"
    notch_depth: float = 0.5
    "The depth of the notch for the switch's clips"
    notch_height: float = 8
    "The height of the notch for the switch's clips at the deepest part of the notch"
    notch_height_outer: float = 9
    "The height of the notch for the switch's clips at the edge of the mounting hole"
    backplate_thickness: float = 1.25
    "The thickness of the backplate, for plate methods that use it"

    backplate_clearance_distance: float = 3.5
    "Depth to clear behind the backplate"
    wall_thickness: float = 1.5
    "The thickness of the mounting hole walls"

    switch_midline_width: float = 13.95
    "The width (left to right) of the switch at its midline (at the top of the plate)"
    switch_midline_length: float = 15.6
    "The length (front to back) of the switch at its midline (at the top of the plate)"
    switch_topline_width: float = 10.5
    "The width (left to right) of the switch at its top"
    switch_topline_length: float = 10.5
    "The length (front to back) of the switch at its top"
    switch_height_above_plate: float = 6.2
    "The height of the switch body above the top of the plate"

    def mx_backplate(self) -> OpenSCADObject:
        """Build a backplate for MX-style switches, with Kailh MX hot-swap socket support.

        You can still use this without hot-swap sockets; the holes are left at the larger size to avoid melting when
        soldering directly.

        See page 10 of https://www.kailhswitch.com/Content/upload/pdf/201815927/PaleBlueSpeed.pdf?rnd=782 for standard
        switch layout, and https://www.kailhswitch.com/Content/upload/pdf/202215927/CPG151101S11-16.pdf?rnd=494 for
        hot-swap socket dimensions.
        """
        return (
            cube((self.keyswitch_width + 3, self.keyswitch_length + 3, self.backplate_thickness), center=True)
            # Center post:
            - cylinder_outer(r=1.995, h=self.backplate_thickness + 1, center=True)
            # Side posts:
            - cylinder_outer(r=0.85, h=self.backplate_thickness + 1, center=True).right(5.08)
            - cylinder_outer(r=0.85, h=self.backplate_thickness + 1, center=True).left(5.08)
            # Contacts:
            - cylinder_outer(r=1.5, h=self.backplate_thickness + 1, center=True).forward(2.54).left(3.81)
            - cylinder_outer(r=1.5, h=self.backplate_thickness + 1, center=True).forward(5.08).right(2.54)
            # LEDs (for up to 4-lead through-hole LEDs):
            - cylinder_outer(r=0.5, h=self.backplate_thickness + 1, center=True).back(5.08).right(1.27)
            - cylinder_outer(r=0.5, h=self.backplate_thickness + 1, center=True).back(5.08).left(1.27)
            - cylinder_outer(r=0.5, h=self.backplate_thickness + 1, center=True).back(5.08).right(3.81)
            - cylinder_outer(r=0.5, h=self.backplate_thickness + 1, center=True).back(5.08).left(3.81)
        ).up(self.plate_thickness - self.keyswitch_depth - self.backplate_thickness / 2)

    def backplate_clearance(self) -> OpenSCADObject:
        """Build a shape to subtract in order to provide clearance around the backplate of the switch.
        """
        z_offset = self.plate_thickness - self.keyswitch_depth - self.backplate_thickness
        return forward(7.5 / 2)(
            up(z_offset - self.backplate_clearance_distance / 2)(
                cube((16.5, 7.5, self.backplate_clearance_distance))
            )
        )

    def plate_with_backplate(self) -> OpenSCADObject:
        """Build a segment of plate for an MX-compatible switch, with a backplate capable of holding a hot-swap socket.
        """
        return self.plate(full_depth=True, extra_depth=1) + self.mx_backplate()


# To test, use the command line: pipenv run python -m spkb.keyswitch.mx
if __name__ == "__main__":
    print("Rendering MX().plate_with_backplate() to mx_plate_with_backplate.scad...")
    MX().plate_with_backplate().save_as_scad("mx_plate_with_backplate.scad")



from typing import Tuple

from solid2 import rotate, cube, up, left, right, forward, back
from solid2.core.object_base import OpenSCADObject

from .switch_plate import (
    switch_plate,
    keyswitch_depth,
    plate_thickness,
    mount_width,
    mount_length,
)


switch_spacing = 2

default_wall_height = keyswitch_depth + 15
wall_length = mount_length + 2 * switch_spacing
wall_thickness = 3


def spaced_switch_plate() -> OpenSCADObject:
    plate_spacer = up(plate_thickness / 2)(
        forward((max(mount_width, mount_length) + switch_spacing) / 2)(
            cube((wall_length, switch_spacing, plate_thickness), center=True)
        )
    )

    return (
        switch_plate()
        + plate_spacer
        + rotate(90)(plate_spacer)
        + rotate(180)(plate_spacer)
        + rotate(270)(plate_spacer)
    )


def key_grid_tester_wall_dimensions(
    length_units: int, width_units: int, wall_height: float, margin_length: float, margin_width: float
) -> Tuple[float, float]:
    wall_length = (
        (mount_length + switch_spacing) * length_units
        + switch_spacing
        + 2 * margin_length
    )
    wall_width = (
        (mount_width + switch_spacing) * width_units + switch_spacing + 2 * margin_width
    )

    return wall_length, wall_width


def key_grid_tester_walls(
    length_units: int, width_units: int, wall_height: float, margin_length: float, margin_width: float
) -> OpenSCADObject:
    wall_length, wall_width = key_grid_tester_wall_dimensions(
        length_units, width_units, wall_height, margin_length, margin_width
    )

    top_wall = forward((wall_length - wall_thickness) / 2)(
        cube((wall_width, wall_thickness, wall_height), center=True)
    )

    left_wall = left((wall_width - wall_thickness) / 2)(
        cube((wall_thickness, wall_length, wall_height), center=True)
    )

    return up(wall_height / 2)(
        top_wall + left_wall + rotate(180)(top_wall) + rotate(180)(left_wall)
    )


def key_grid_tester(
    length_units: int,
    width_units: int,
    wall_height: float = default_wall_height,
    margin_length: float = 0,
    margin_width: float = 0,
) -> OpenSCADObject:
    x_grid_size = mount_width + switch_spacing
    y_grid_size = mount_length + switch_spacing

    case = key_grid_tester_walls(
        length_units, width_units, wall_height, margin_length, margin_width
    ) + up(wall_height - plate_thickness)(
        right(x_grid_size * (width_units - 1) / 2)(
            back(y_grid_size * (length_units - 1) / 2)(
                *[
                    left(x_grid_size * x_units)(
                        forward(y_grid_size * y_units)(spaced_switch_plate())
                    )
                    for y_units in range(length_units)
                    for x_units in range(width_units)
                ]
            )
        )
    )

    return case



"""Utility functions for building things with SolidPython2.
"""
from collections.abc import Callable, Sequence
from math import pi, cos
from typing import List, Union

from solid2 import cube, cylinder
from solid2.core.object_base import OpenSCADObject


def fudge_radius(r: Union[float, Sequence[float]], segments: int = 16) -> Union[float, List[float]]:
    """Adjust the given radius for the given number of segments to make it generate a circumscribed circular object.

    See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/undersized_circular_objects for more info.

    :param r: The radius of the cylinder, or a sequence containing the radii of the top and bottom ends.
    :param segments: Number of fragments in 360 degrees.
    """
    fudge = 1 / cos(pi / segments)
    return [ri * fudge for ri in r] if isinstance(r, Sequence) else r * fudge


def cylinder_outer(r: Union[float, Sequence[float]], h: float = 1, segments: int = 16, center: bool = False) -> OpenSCADObject:
    """Create a cylinder using circumscribed polygons instead of the default inscribed polygons.

    See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/undersized_circular_objects for more info.

    :param r: The radius of the cylinder, or a sequence containing the radii of the top and bottom ends.
    :param h: This is the height of the cylinder.
    :param segments: Number of fragments in 360 degrees.
    :param center: If True will center the height of the cone/cylinder around
    the origin. Default is False, placing the base of the cylinder or r1 radius
    of cone at the origin.
    """
    adjusted_r = fudge_radius(r, segments)

    radii = {}
    if isinstance(adjusted_r, (int, float)):
        radii['r'] = adjusted_r
    elif isinstance(adjusted_r, (tuple, list)):
        radii['r1'], radii['r2'] = adjusted_r

    return cylinder(
        h=h,
        _fn=segments,
        center=center,
        **radii,
    )


nothing = cube((1, 1, 1), center=True) - cube((2, 2, 2), center=True)
"Nothing. (a completely empty shape)"


def optional(condition: bool) -> Callable[[OpenSCADObject], OpenSCADObject]:
    """Optionally include the wrapped part.

    If `condition` is truthy, the wrapped part will be returned; otherwise, spkb.utils.nothing will be returned.

    :param condition: the condition under which to include the part
    """
    return lambda part: part if condition else nothing



# To test, use the command line: pipenv run python -m spkb.keyswitch

from . import mx_keyswitch

print("Rendering mx_keyswitch() to mx_keyswitch.scad...")
mx_keyswitch().save_as_scad("mx_keyswitch.scad")



"""Classes representing different types of keyswitches, able to generate switch plates, backplates, etc.
"""
import sys
from typing_extensions import deprecated

from solid2 import cube, hull
from solid2.core.object_base import OpenSCADObject

if not "-m" in sys.argv:
    from .choc import Choc
    from .mx import MX

    __all__ = [
        "Choc",
        "MX",
        "mx_midline_width", "mx_midline_length", "mx_topline_width", "mx_topline_length", "mx_height_above_plate",
        "mx_keyswitch",
    ]


mx_midline_width = 13.95
mx_midline_length = 15.6
mx_topline_width = 10.5
mx_topline_length = 10.5
mx_height_above_plate = 6.2


@deprecated("Use MX().keyswitch() instead")
def mx_keyswitch() -> OpenSCADObject:
    """Build an simplified approximation of (the top half of) an MX-style keyswitch.
    """
    return hull()(
        cube(mx_midline_width, mx_midline_length, 0.1, center=True).up(3.05),
        cube(mx_topline_width, mx_topline_length, 0.1, center=True).up(mx_height_above_plate + 2.95),
    )



"""Renderers for approximations of keycap shapes.
"""
from solid2 import polygon, hull
from solid2.core.object_base import OpenSCADObject

from spkb.switch_plate import plate_thickness


sa_length = 18.25
sa_1_5_length = 28
sa_double_length = 37.5
sa_mid_shrink = 0.625
sa_top_shrink = 3.125


def sa_cap(units: float) -> OpenSCADObject:
    """Create an "SA" keycap shape.

    (actually something akin to SA row 3, but with flat tops)

    :param units: the size of the key, in key units
    """
    if units == 1:
        bl2 = sa_length / 2
        ml2 = sa_length / 2 - sa_mid_shrink
        tl2 = sa_length / 2 - sa_top_shrink
        key_cap = hull()(
            polygon(((bl2, bl2), (bl2, -bl2), (-bl2, -bl2), (-bl2, bl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 0.05)),
            polygon(((ml2, ml2), (ml2, -ml2), (-ml2, -ml2), (-ml2, ml2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 6)),
            polygon(((tl2, tl2), (tl2, -tl2), (-tl2, -tl2), (-tl2, tl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 12)),
        )
        return (
            key_cap
            .translate((0, 0, 5 + plate_thickness))
            .color((220/255, 163/255, 163/255, 1))
        )

    elif units == 2:
        bl2 = sa_double_length / 2
        bw2 = sa_length / 2
        ml2 = sa_double_length / 2 - sa_mid_shrink
        mw2 = sa_length / 2 - sa_mid_shrink
        tl2 = sa_double_length / 2 - sa_top_shrink
        tw2 = sa_length / 2 - sa_top_shrink
        key_cap = hull()(
            polygon(((bw2, bl2), (bw2, -bl2), (-bw2, -bl2), (-bw2, bl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 0.05)),
            polygon(((mw2, ml2), (mw2, -ml2), (-mw2, -ml2), (-mw2, ml2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 6)),
            polygon(((tw2, tl2), (tw2, -tl2), (-tw2, -tl2), (-tw2, tl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 12)),
        )
        return (
            key_cap
            .translate((0, 0, 5 + plate_thickness))
            .color((127/255, 159/255, 127/255, 1))
        )

    elif units == 1.5:
        bl2 = sa_1_5_length / 2
        bw2 = sa_length / 2
        ml2 = sa_1_5_length / 2 - sa_mid_shrink
        mw2 = sa_length / 2 - sa_mid_shrink
        tl2 = sa_1_5_length / 2 - sa_top_shrink
        tw2 = sa_length / 2 - sa_top_shrink
        key_cap = hull()(
            polygon(((bw2, bl2), (bw2, -bl2), (-bw2, -bl2), (-bw2, bl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 0.05)),
            polygon(((mw2, ml2), (mw2, -ml2), (-mw2, -ml2), (-mw2, ml2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 6)),
            polygon(((tw2, tl2), (tw2, -tl2), (-tw2, -tl2), (-tw2, tl2)))
            .linear_extrude(height=0.1, twist=0, convexity=0)
            .translate((0, 0, 12)),
        )
        return (
            key_cap
            .translate((0, 0, 5 + plate_thickness))
            .color((240/255, 223/255, 175/255, 1))
        )

    else:
        raise ValueError(f"Unrecognized key size: {units} units")


# To test, use the command line: pipenv run python -m spkb.keycaps
if __name__ == "__main__":
    print("Rendering sa_cap(1), sa_cap(1.5), and sa_cap(2) to keycaps.scad...")
    (
        sa_cap(1).translate((-30, 0, 0))
        + sa_cap(1.5)
        + sa_cap(2).translate((30, 0, 0))
    ).save_as_scad("keycaps.scad")
